from os.path import abspath, exists, dirname
from os import makedirs
from glob import glob
from importlib import import_module

import cherrypy

from yaml import load, add_constructor

from .sloth import Sloth


class Bed:
    '''A container for Sloth apps.

    It tracks their config files and listen points, as well as handles app adding, removing, and updating.

    If also implements the listen function for the main CherryPy app.

    (This module is names "bed" because a group of sloths is actually called "bed.")
    '''

    def __new__(cls, config):
        '''Apply extenions before creating a Bed instance.

        The built-in API extension is always applied before any custom ones.

        :param config: bed config dict
        '''

        api_extension = {
            'api': {
                'module': 'api.bed'
            }
        }

        PreExtendedBed, pre_errors = cls.extend(api_extension)

        ExtendedBed, errors = PreExtendedBed.extend(config.get('extensions'))

        for error in pre_errors + errors:
            cherrypy.log.error(error)

        return super().__new__(ExtendedBed)

    def __init__(self, config):
        '''Configure CherryPy loop to listen for payload.

        :param config: bed config dict
        '''

        self.config = config
        self.bus = cherrypy.engine

        self.sloths = {}
        self.config_files = {}

        self._configure()
        self._prepopulate()


    @staticmethod
    def _critical_constructor(loader, node):
        '''Custom PyYAML constructor for the ``!critical`` tag.

        :param loader: PyYAML ``Loader`` instance
        :param node: PyYAML ``Node`` instance

        :returns: :class:`Action` instance
        '''

        class Action(str):
            '''Thin wrapper around ``str`` to enable custom attributes.

            Used to declare actions as *critical*.
            '''
            pass

        action = Action(node.value)
        action.critical = True

        return action

    def _configure(self):
        '''Configure CherryPy server.'''

        self.db_path = self.config.get('paths', {}).get('db', 'sloth.db')

        if self.db_path:
            db_dir = dirname(abspath(self.db_path))

            if db_dir and not exists(db_dir):
                makedirs(db_dir)

            self.db_extensions = {
                'app_logs': {
                    'module': 'db.logs',
                    'db': self.db_path,
                    'table': 'app_logs'
                },
                'build_history': {
                    'module': 'db.builds',
                    'db': self.db_path,
                    'table': 'build_history'
                }
            }

        access_log_path = self.config.get('paths', {}).get('access_log', 'sloth_access.log')
        access_log_dir = dirname(access_log_path)

        if access_log_dir and not exists(access_log_dir):
            makedirs(access_log_dir)

        error_log_path = self.config.get('paths', {}).get('error_log', 'sloth_error.log')
        error_log_dir = dirname(error_log_path)

        if error_log_dir and not exists(error_log_dir):
            makedirs(error_log_dir)

        cherrypy.config.update(
            {
                'environment': 'production',
                'server.socket_host': self.config['host'],
                'server.socket_port': self.config['port'],
                'log.access_file': access_log_path,
                'log.error_file': error_log_path
            }
        )

        self._setup_routing()

        cherrypy.tree.mount(None, config={'/': {'request.dispatch': self._dispatcher}})

        if self.config.get('daemon'):
            cherrypy.process.plugins.Daemonizer(self.bus).subscribe()

        self.bus.subscribe('stop', self.remove_all)

    def _setup_routing(self):
        '''Add routes for app listener.'''

        self._dispatcher = cherrypy._cpdispatch.RoutesDispatcher()

        self._dispatcher.connect('apps', '/{listen_point:.+}', self.listener)

    def _prepopulate(self):
        '''Create apps before server start.

        The app configs are extracted from the files defined in the config_paths section of the server config.
        '''

        for config_path in self.config.get('paths', {}).get('configs', []):
            config_files = glob(config_path)

            if not config_files:
                cherrypy.log.error('Path %s not found' % config_path)
                continue

            for config_file in config_files:
                try:
                    add_constructor('!critical', self._critical_constructor)

                    config = load(open(config_file))

                    listen_point = self.create_from_config(config)

                    self.bind_to_file(listen_point, abspath(config_file))

                except:
                    continue

    @classmethod
    def extend(cls, extensions):
        '''Sequentially chain-inherit Bed classes from extensions.

        The first extension's Bed class inherits from the base Bed class and becomes the base class, then the second one inherits from it, and so on.

        :param extensions: list of extensions to load.

        :returns: `ExtendedBed` is a Bed class inherited from all extensions' Bed classes; `errors` is the list of errors raised during extension loading.
        '''

        ExtendedBed = cls
        errors = []

        if extensions:
            for extension_name, extension_config in extensions.items():
                try:
                    ext = import_module('.ext.%s' % extension_config['module'], package=__package__)

                    ExtendedBed = ext.extend_bed(ExtendedBed, {
                            'name': extension_name,
                            'config': extension_config
                        }
                    )

                except AttributeError as e:
                    pass

                except Exception as e:
                    errors.append('Could not load extension %s: %s' % (extension_name, e))

        return ExtendedBed, errors

    def start(self):
        '''Start CherryPy loop to listen for payload.'''

        self.bus.start()
        self.bus.block()

    def bind_to_file(self, listen_point, config_file):
        '''Bind a Sloth app with a config file.

        :param listen_point: app's listen point
        :param config_file: absolute path to the config file
        '''

        try:
            sloth = self.sloths[listen_point]

            if sloth.config == load(open(config_file)):
                self.config_files[listen_point] = config_file
                sloth.logger.info('Bound with config file %s' % config_file)

                cherrypy.log.error('App on %s bound with config file %s' % (listen_point, config_file))

            else:
                raise ValueError

        except KeyError:
            cherrypy.log.error('Failed to bind config file: listen point %s not found' % listen_point)
            raise

        except FileNotFoundError:
            cherrypy.log.error('Failed to bind config file: file %s not found' % config_file)
            raise

        except ValueError:
            cherrypy.log.error('Failed to bind config file: config mismatch')
            raise

    def create_from_config(self, config):
        '''Create a Sloth app from a config source and add it to the bed.

        :param config: a dict parsed from YAML

        :returns: new app's listen point
        '''

        try:
            for alias in ('id', 'name', 'listen_point'):
                listen_point = config.get(alias)
                if listen_point:
                    break
            else:
                cherrypy.log.error('Failed to create app: listen point undefined')
                raise KeyError('id')

        except Exception as e:
            cherrypy.log.error('Failed to create app: invalid config: %s' % e)
            raise

        if listen_point in self.sloths:
            cherrypy.log.error('Failed to create app: listen point %s already taken' % listen_point)
            raise ValueError(listen_point)

        try:
            if self.db_path:
                PreExtendedSloth, pre_errors = Sloth.extend(self.db_extensions)

            else:
                PreExtendedSloth = Sloth
                pre_errors = []

            ExtendedSloth, errors = PreExtendedSloth.extend(config.get('extensions'))
            sloth = ExtendedSloth(config)

            for error in pre_errors + errors:
                sloth.logger.error(error)

            self.sloths[listen_point] = sloth
            sloth.logger.info('Listening on %s' % listen_point)

            cherrypy.log.error('Sloth app added, listening on %s' % listen_point)

            return listen_point

        except Exception as e:
            cherrypy.log.error('Failed to create app: %s' % e)
            raise

    def remove(self, listen_point):
        '''Stop Sloth app and remove it from the bed.

        :param listen_point: Sloth app listen point
        '''

        try:
            self.sloths.pop(listen_point).stop()

            self.config_files.pop(listen_point, None)

            cherrypy.log.error('Sloth app at %s removed' % listen_point)

        except KeyError:
            cherrypy.log.error('Failed to remove app: listen point %s not found' % listen_point)
            raise

        except Exception as e:
            cherrypy.log.error('Failed to remove app on %s: %s' % (listen_point, e))

    def remove_all(self):
        '''Stop all active Sloth apps and remove them from the bed.'''

        while self.sloths:
            listen_point, sloth = self.sloths.popitem()

            sloth.stop()

            self.config_files.pop(listen_point, None)

    @cherrypy.expose
    @cherrypy.tools.proxy()
    @cherrypy.tools.json_in()
    def listener(self, listen_point, **kwargs):
        '''Listens for payloads and routes them to the responsible Sloth app.

        :param listen_point: Sloth app listen point (part of the URL after the server host)
        '''

        sloth = self.sloths.get(listen_point)

        if sloth:
            sloth.handle(cherrypy.request)

        else:
            raise cherrypy.HTTPError(404, 'This listen point does not exist.')



def extend_cli(cls, extension):
    '''Add command-line commands for the API interaction.

    :param cls: base :class:`CLI <sloth_ci.cli.CLI>` class to be extended
    :param extension: ``{'api': {'module': 'api'}}``

    :returns: extended :class:`CLI <sloth_ci.cli.CLI>` class
    '''

    from sys import exit
    from os.path import abspath
    from glob import glob
    from collections import namedtuple
    from functools import partial
    from time import ctime

    from tabulate import tabulate
    from cliar import add_aliases

    from requests import post, exceptions


    class CLI(cls):
        '''Welcome to Sloth CI API CLI!

Run "sci start" to start the server.

Run "sci -h" to see all available commands.

Run "sci <command> -h" to get help for a specific command.
'''

        def __init__(self):
            super().__init__()

            api_url = 'http://%s:%d' % (self.config['host'], self.config['port'])

            for alias in ('auth', 'api_auth'):
                auth = self.config.get(alias)
                if auth:
                    api_auth = (auth['login'], auth['password'])
                    break
            else:
                api_auth = None

            self.send_api_request = partial(self.send_api_request, api_url, api_auth)

        @staticmethod
        def send_api_request(url, auth, data={}):
            '''Send a POST request to a Sloth CI API server with the given data.

            :param url: URL of the Sloth CI API server
            :param auth: login and password to access the Sloth CI API
            :param data: dict of data to be sent with the request
            '''

            APIResponse = namedtuple('APIResponse', ('status_code', 'content'))

            try:
                response = post(url, auth=auth, data=data)

                if response.ok:
                    if response.content:
                        content = response.json()

                    else:
                        content = None

                else:
                    content = response.text.strip()

                return APIResponse(response.status_code, content)

            except exceptions.ConnectionError:
                return APIResponse(503, 'Failed to connect to Sloth CI on %s' % url)

        @staticmethod
        def colorize(table, based_on_column, hide_level=True):
            '''Colorize logs table according to message status.

            :param table: logs table, each row is [timestamp, message, level_name]
            :param based_on_column: number of column that holds record level names for coloring

            :returns: colorized table
            '''
            try:
                from colorama import init, Fore, Back

                init()
                colorized_table = []

            except ImportError:
                colorized_table = table

            if not colorized_table:
                for row in table:
                    reset = Fore.RESET + Back.RESET

                    level_name = row[based_on_column]

                    if level_name == 'DEBUG':
                        color = Fore.CYAN
                    elif level_name == 'INFO':
                        color = Fore.GREEN
                    elif level_name == 'WARNING':
                        color = Fore.YELLOW
                    elif level_name == 'ERROR':
                        color = Fore.RED
                    elif level_name == 'CRITICAL':
                        color = Fore.RED + Back.WHITE
                    else:
                        color = reset

                    colorized_row = [(color + cell + reset) for cell in row]
                    colorized_table.append(colorized_row)

            if hide_level:
                for row in colorized_table:
                    row.pop(based_on_column)

            return colorized_table


        @add_aliases(['add'])
        def create(self, paths:tuple):
            '''create apps from PATHS and bind them with respective configs'''

            for path in paths:
                config_files = glob(path)

                if not config_files:
                    print('Path %s not found' % path)
                    continue

                for config_file in config_files:
                    response = self.send_api_request(
                        {
                            'action': 'create',
                            'config_string': open(config_file).read()
                        }
                    )

                    if response.status_code == 201:
                        listen_point = response.content
                        print('App "%s" created' % listen_point)

                    else:
                        print('Failed to create app: %s' % response.content)
                        continue

                    response = self.send_api_request(
                        {
                            'action': 'bind',
                            'listen_point': listen_point,
                            'config_file': abspath(config_file)
                        }
                    )

                    if response.status_code == 200:
                        print('App "%s" bound with config file "%s"' % (listen_point, config_file))

                    else:
                        print('Failed to bind app with config file: %s' % response.content)

        @add_aliases(['hist', 'builds'])
        def history(self, app, from_page:int=1, to_page:int=1, per_page:int=10, level:int=20, verbose=False):
            '''get build history for APP'''

            response = self.send_api_request(
                {
                    'action': 'history',
                    'listen_point': app,
                    'from_page': from_page,
                    'to_page': to_page,
                    'per_page': per_page,
                    'level': level
                }
            )

            if response.status_code == 200:
                rows = [
                    [
                        ctime(record['timestamp']),
                        record['message'],
                        record['level_name']
                    ] for record in response.content
                ]

                headers = [
                    'Timestamp',
                    'Status',
                    'Level'
                ]

                if verbose:
                    table = tabulate(
                        self.colorize(rows, based_on_column=-1, hide_level=False),
                        headers=headers
                    )

                else:
                    table = tabulate(
                        self.colorize(rows, based_on_column=-1),
                        headers=headers
                    )

                print(table)

            else:
                exit('Failed to get app build history: %s' % response.content)

        def info(self, app):
            '''get information about APP'''

            response = self.send_api_request(
                {
                    'action': 'info',
                    'listen_point': app
                }
            )

            if response.status_code == 200:
                info = response.content

                rows = [[
                    info['config_file'],
                    info['last_build_status_message'],
                    ctime(info['last_build_timestamp']),
                    info['last_build_status_level']
                ]]

                headers = [
                    'Config File',
                    'Last Build Message',
                    'Last Build Timestamp'
                ]

                table = tabulate(
                    self.colorize(rows, based_on_column=-1),
                    headers=headers
                )

                print(table)

            else:
                exit('Failed to get app info: %s' % response.content)

        @add_aliases(['ls'])
        def list(self):
            '''list all available apps'''

            response = self.send_api_request(
                {
                    'action': 'list'
                }
            )

            if response.status_code == 200:
                for app in response.content:
                    print(app)

            else:
                exit('Failed to get app list: %s' % response.content)

        @add_aliases(['lg'])
        def logs(self, app, from_page:int=1, to_page:int=1, per_page:int=10, level:int=20, verbose=False):
            '''get logs for APP'''

            response = self.send_api_request(
                {
                    'action': 'logs',
                    'listen_point': app,
                    'from_page': from_page,
                    'to_page': to_page,
                    'per_page': per_page,
                    'level': level
                }
            )

            if response.status_code == 200:
                rows = [
                    [
                        ctime(record['timestamp']),
                        record['message'],
                        record['level_name']
                    ] for record in response.content
                ]

                headers = [
                    'Timestamp',
                    'Message',
                    'Level'
                ]

                if verbose:
                    table = tabulate(
                        self.colorize(rows, based_on_column=-1, hide_level=False),
                        headers=headers
                    )

                else:
                    table = tabulate(
                        self.colorize(rows, based_on_column=-1),
                        headers=headers
                    )

                print(table)

            else:
                exit('Failed to get app logs: %s' % response.content)

        @add_aliases(['update', 'up'])
        def reload(self, app):
            '''recreate APP to apply config changes'''

            response = self.send_api_request(
                {
                    'action': 'info',
                    'listen_point': app
                }
            )

            if response.status_code == 200:
                self.remove(app)
                self.create([response.content['config_file']])

            else:
                exit('Failed to get app info: %s' % response.content)

        @add_aliases(['del', 'rm'])
        def remove(self, app):
            '''remove APP'''

            response = self.send_api_request(
                {
                    'action': 'remove',
                    'listen_point': app
                }
            )

            if response.status_code == 204:
                print('App "%s" removed' % app)

            else:
                exit('Failed to remove app: %s' % response.content)

        def restart(self):
            '''restart server'''

            response = self.send_api_request(
                {
                    'action': 'restart'
                }
            )

            if response.status_code == 202:
                print('Restarting Sloth CI on http://%s:%d' % (self.config['host'], self.config['port']))

            else:
                exit('Failed to restart Sloth CI: %s' % response.content)

        @add_aliases(['stat', 'st'])
        def status(self):
            '''check server status'''

            response = self.send_api_request(
                {
                    'action': 'version'
                }
            )

            if response.status_code == 200:
                print(
                    'Sloth CI version %s is running on http://%s:%d'
                     % (response.content, self.config['host'], self.config['port'])
                )

            else:
                print('Sloth CI is not running on http://%s:%d' % (self.config['host'], self.config['port']))

        def stop(self):
            '''stop server'''

            response = self.send_api_request(
                {
                    'action': 'stop'
                }
            )

            if response.status_code == 202:
                print('Stopping Sloth CI on http://%s:%d' % (self.config['host'], self.config['port']))

            else:
                exit('Failed to stop Sloth CI: %s' % response.content)


        @add_aliases(['run', 'fire'])
        def trigger(self, app, wait=False, params=()):
            '''trigger APP's actions with given PARAMS

PARAMS are specified as "param1=value1 param2=value ..."
'''

            data = {
                'action': 'trigger',
                'listen_point': app,
                'wait': wait or ''
            }

            for param in params:
                key, value = param.split('=')
                data[key] = value

            response = self.send_api_request(data)

            if response.status_code == 202:
                print('Actions triggered on %s' % app)

            elif response.status_code == 200:
                print(response.content)

            else:
                exit('Failed to trigger actions: %s' % response.content)


    return CLI



def extend_sloth(cls, extension):
    '''Add SQLite logging for all app events.

    :param cls: base :mod:`Sloth <sloth_ci.sloth.Sloth>` class to extend
    :param extension: ``{'app_logs': {'module': 'db.logs', 'db': self.db_path, 'table': 'app_logs'}``

    :returns: extended :mod:`Sloth <sloth_ci.sloth.Sloth>` class
    '''

    import logging

    from .util import SqliteHandler


    class Sloth(cls):
        def __init__(self, config):
            super().__init__(config)

            db_app_log_config = extension['config']

            db_app_log_handler = SqliteHandler(db_app_log_config['db'], db_app_log_config['table'])

            db_app_log_handler.setLevel(logging.DEBUG)

            self.logger.addHandler(db_app_log_handler)

            self.log_handlers[extension['name']] = db_app_log_handler

        def stop(self):
            super().stop()
            self.logger.removeHandler(self.log_handlers.pop(extension['name']))


    return Sloth
